/*
 * Some css adaptions to the Furo theme for parce.info.
 */


/* add extra spacing above class and function definitions */
@media screen {
    div.section > dl.class {
        margin-top: 40px;
    }

    div.section > dl.function {
        margin-top: 30px;
    }
}

